// This file is part of retro-gtk. License: GPL-3.0+.

#pragma once

#define __RETRO_GTK_INSIDE__

#ifndef RETRO_GTK_USE_UNSTABLE_API
#error    retro-gtk is unstable API. You must define RETRO_GTK_USE_UNSTABLE_API before including retro-gtk.h
#endif

#include "retro-controller.h"
#include "retro-controller-codes.h"
#include "retro-controller-iterator.h"
#include "retro-controller-type.h"
#include "retro-core.h"
#include "retro-core-descriptor.h"
#include "retro-core-view.h"
#include "retro-gtk-version.h"
#include "retro-input.h"
#include "retro-key-joypad-mapping.h"
#include "retro-log.h"
#include "retro-memory-type.h"
#include "retro-module-iterator.h"
#include "retro-module-query.h"
#include "retro-option.h"
#include "retro-option-iterator.h"
#include "retro-pixbuf.h"
#include "retro-pixdata.h"
#include "retro-rumble-effect.h"
#include "retro-video-filter.h"

#undef __RETRO_GTK_INSIDE__
